clc;
clear all;

%% Noise distribution

figure(1); 
D = abs(randn(20000, 1));

histogram(D, 60, 'Normalization', 'probability');

%% Model
M = 80;   % total number of timesteps
N = 15;   % horizon length

nx = 2;
ny = 1;

% system dynamics and output function 
f = @(x) [0.99*x(1) + 0.3*x(2);
         -0.05*x(1) + 0.9*sin(x(2));];

h = @(x) x(1) - 3*x(2);

% Jacobians
J_f = @(x) [0.99, 0.3;
            -0.05, 0.9*cos(x(2));];
J_h = @(x) [1, -3];

Q = 0.05*eye(nx);    % state noise covariance
R = 0.05*eye(ny);    % measurement noise covariance

x0 = [0.8; 0.8];     % prior mean
Q0 = eye(nx);        % prior covariance

%% Load data

load('data_nonlin.mat');

% plot the true state and the measurements
figure(2);
subplot(2, 1, 1); grid on; hold on;
plot(1:M, X(1, :), '.-');
plot(1:M, X(2, :), '.-');
legend('x1', 'x2');
xlabel('timestep k');

subplot(2, 1, 2); grid on; hold on;
plot(1:M, Y(1, :), '.-');
legend('y1');
xlabel('timestep k');


%% Kalman Filter

x_estimate_kf = zeros(nx, M);

x_ = x0;
P_ = Q0;

for k=1:M

    [x, P] = update_ekf(Y(:, k), x_, P_, h, J_h, R);
    [x_, P_] = predict_ekf(x, P, f, J_f, Q);
    
    x_estimate_kf(:, k) = x;
end


%% Moving Horizon Estimation

x_estimate_mhe_uc = zeros(nx, M-N);
x_estimate_mhe_c = zeros(nx, M-N);

% forward simulation for initial guess
x_init = zeros(nx, N);
x_init(:, 1) = x0;

for k=1:N-1
    x_init(:, k+1) = f(x_init(:, k)) + 0.1;
end


%% Unconstrained
for k=1:M-N
    Y_current = Y(:, k:k+N);
    
    % cost with zero arrival cost term
    cost = @(X) objective(X, Y_current, f, h, Q, R, [], [], false);
    
    x_opt = fmincon(cost, x_init);
   
    x_estimate_mhe_uc(:, k) = x_opt(:, end);
    
    x_init(:, 1:N-1) = x_opt(:, 2:N);
    x_init(:, N) = f(x_init(:, N-1));
end

%% Constrained
constr = @(X) constraints(X, f);

opt = optimoptions('fmincon', 'MaxFunctionEvaluations', 5000);

for k=1:M-N

    Y_current = Y(:, k:k+N);
    cost = @(X) objective(X, Y_current, f, h, Q, R, [], [], false);
    
    x_opt = fmincon(cost, x_init, [], [], [], [], [],[], constr, opt);
   
    x_estimate_mhe_c(:, k) = x_opt(:, end);
    
    x_init(:, 1:N-1) = x_opt(:, 2:N);
    x_init(:, N) = f(x_init(:, N-1));
end

%% Estimation results

figure(3); 

subplot(2,1,1); hold on; grid on;

plot(1:M, X(1,:), '-');
plot(1:M, x_estimate_kf(1,:), '.-');
plot(N:M-1, x_estimate_mhe_uc(1,:), '.-');
plot(N:M-1, x_estimate_mhe_c(1,:), '.-');
legend('true', 'EKF', 'MHE unconstrained', 'MHE constrained');

subplot(2,1,2); hold on; grid on;

plot(1:M, X(2,:), '-');
plot(1:M, x_estimate_kf(2,:), '.-');
plot(N:M-1, x_estimate_mhe_uc(2,:), '.-');
plot(N:M-1, x_estimate_mhe_c(2,:), '.-');

legend('true', 'EKF', 'MHE unconstrained', 'MHE constrained');


save('data_nonlin.mat', 'X', 'Y');
